/* 
 * Akonadi OpenSync 0.22 plugin
 * Copyright (C) 2010  Tejas Guruswamy <masterpatricko@gmail.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <stdlib.h>
#include <assert.h>
#include <dlfcn.h>

#include <QXmlStreamReader>
#include <QLibrary>
#include <kdemacros.h>

#include "plugin.h"
#include "akonadisync_wrapper.h"

extern "C" {
    static void* initialize(OSyncMember *member, OSyncError **error)
    {
        osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);

        //You need to specify the <some name>_environment somewhere with
        //all the members you need
        plugin_environment *env = (plugin_environment*)osync_try_malloc0(sizeof(plugin_environment), error);
        assert(env != NULL);
        memset(env, 0, sizeof(plugin_environment));
        
        //now you can get the config file for this plugin
        char *configdata;
        int configsize = 0;
        
        if (!osync_member_get_config(member, &configdata, &configsize, error)) {
            osync_error_update(error, "Unable to get config data: %s", osync_error_print(error));
            free(env);
            return NULL;
        }
        
        //Process the configdata here and set the options on your environment
        QXmlStreamReader reader;
        qint64 contact_collection = 0;
        qint64 event_collection = 0;
        bool ok = true;
        reader.addData( QByteArray::fromRawData(configdata, configsize) );
        while ( !reader.atEnd() ) {
            reader.readNext();
            if ( reader.isStartElement() && reader.name() == "config" )
                continue;
            if ( reader.isStartElement() && reader.name() == "contacts" ) {
                QString text = reader.readElementText();
                if ( !text.isEmpty() )
                    contact_collection = text.toLongLong(&ok); 
            }
            if ( reader.isStartElement() && reader.name() == "events" ) {
                QString text = reader.readElementText();
                if ( !text.isEmpty() )
                    event_collection = text.toLongLong(&ok);
            }
        }
        
        if ( reader.hasError() || !ok ) {
            osync_error_set(error, OSYNC_ERROR_MISCONFIGURATION, "Unable to parse config data: %s, %s",
                            ok ? "true" : "false" , reader.errorString().toAscii().constData());
            free(configdata);
            free(env);
            return NULL;
        }
        
        osync_debug("AKONADI-SYNC", 3, "Read default collections as %lld (contacts) %lld (events)",
                    contact_collection, event_collection);
        
        free(configdata);
        
        env->member = member;
        
        //If you need a hashtable you make it here
        env->hashtable = osync_hashtable_new();
        
        QByteArray modulepath = OPENSYNC_PLUGINDIR"/akonadi_synclib.so";
        osync_debug("AKONADI-SYNC", 3, "Loading implementation module from %s", modulepath.constData());
        QLibrary module(modulepath);
        if (module.load()) {
            osync_error_set(error, OSYNC_ERROR_INITIALIZATION, "Can't load plugin implementation module from %s: %s",
                            modulepath.constData(), module.errorString().toAscii().constData());
        }
        osync_debug("AKONADI-SYNC", 3, "Getting initialization function");
        AkonadiSyncInitFunc init_func = (AkonadiSyncInitFunc)module.resolve("new_AkonadiSync");
        if (!init_func) {
            osync_error_set(error, OSYNC_ERROR_INITIALIZATION, "Invalid plugin implementation module");
        }
        
        osync_debug("AKONADI-SYNC", 3, "Initializing implementation module");
        
        env->sync = init_func(contact_collection, event_collection);
        
        //Now your return your struct.
        osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
        return (void *)env;
    }

    static void connect(OSyncContext *ctx)
    {
        osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);;
        //Each time you get passed a context (which is used to track
        //calls to your plugin) you can get the data your returned in
        //initialize via this call:
        plugin_environment *env = (plugin_environment *)osync_context_get_plugin_data(ctx);

        int enabledtypes = 0;
        if ( osync_member_objtype_enabled(env->member, "contact") ) 
            enabledtypes |= ContactsEnabled;
        if ( osync_member_objtype_enabled(env->member, "event") ) 
            enabledtypes |= EventsEnabled;
        if ( osync_member_objtype_enabled(env->member, "todo") ) 
            enabledtypes |= TodoEnabled;
        if ( osync_member_objtype_enabled(env->member, "note") ) 
            enabledtypes |= NotesEnabled;
        
        osync_debug("AKONADI-SYNC", 2,"Types enabled bitflags set to %d", enabledtypes);
        
        /*
        * Now connect to your devices and report
        * 
        * an error via:
        * osync_context_report_error(ctx, ERROR_CODE, "Some message");
        * 
        * or success via:
        * osync_context_report_success(ctx);
        * 
        * You have to use one of these 2 somewhere to answer the context.
        * 
        */
        if ( !env->sync->initialize(enabledtypes) ) {
            osync_context_report_error(ctx, OSYNC_ERROR_INITIALIZATION, "Couldn't initialize akonadi");
            return;
        }
        
        OSyncError *error = NULL;
        //If you are using a hashtable you have to load it here
        if ( !osync_hashtable_load(env->hashtable, env->member, &error) ) {
            osync_context_report_osyncerror(ctx, &error);
            return;
        }
        
        //you can also use the anchor system to detect a device reset
        //or some parameter change here. Check the docs to see how it works
        //FIXME: Anchor should be a random string unique to each system
        const char *lanchor = "akonadi-sync";
        //Now you get the last stored anchor from the device
        if (!osync_anchor_compare(env->member, "lanchor", lanchor)) {
            osync_member_set_slow_sync(env->member, "contact", TRUE);
            osync_member_set_slow_sync(env->member, "event", TRUE);
            osync_member_set_slow_sync(env->member, "todo", TRUE);
            osync_member_set_slow_sync(env->member, "note", TRUE);
        }
        
        osync_context_report_success(ctx);
        osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
        return;
    }

    static void get_changeinfo(OSyncContext *ctx)
    {
        osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
        plugin_environment *env = (plugin_environment *)osync_context_get_plugin_data(ctx);
        
        //If you use opensync hashtables you can detect if you need
        //to do a slow-sync and set this on the hastable directly
        //otherwise you have to make 2 function like "get_changes" and
        //"get_all" and decide which to use using
        //osync_member_get_slow_sync
        if (osync_member_get_slow_sync(env->member, "contact"))
            osync_hashtable_set_slow_sync(env->hashtable, "contact");
        
        if (osync_member_get_slow_sync(env->member, "event"))
            osync_hashtable_set_slow_sync(env->hashtable, "event");
        
        if (osync_member_get_slow_sync(env->member, "todo"))
            osync_hashtable_set_slow_sync(env->hashtable, "todo");
        
        if (osync_member_get_slow_sync(env->member, "note"))
            osync_hashtable_set_slow_sync(env->hashtable, "note");

        /*
        * Now you can get the changes.
        * Loop over all changes you get and do the following:
        */  
        QList<OSyncChange*> changes = env->sync->listItems();
        foreach ( OSyncChange* change, changes) {
            //Now get the data of this change
            
            //Set the member
            osync_change_set_member(change, env->member);			

            //If you use hashtables use these functions:
            if (osync_hashtable_detect_change(env->hashtable, change)) {
                //There was a change, get the full data
                env->sync->updateItem(change, TRUE);
                osync_context_report_change(ctx, change);
                osync_hashtable_update_hash(env->hashtable, change);
            }
            //otherwise just report the change via
            //osync_context_report_change(ctx, change);
        }

        //When you are done looping and if you are using hashtables	
        osync_hashtable_report_deleted(env->hashtable, ctx, "contact");
        osync_hashtable_report_deleted(env->hashtable, ctx, "event");
        osync_hashtable_report_deleted(env->hashtable, ctx, "todo");
        osync_hashtable_report_deleted(env->hashtable, ctx, "note");
        
        //Now we need to answer the call
        osync_context_report_success(ctx);
        
        osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
        return;
    }

    static osync_bool commit_change(OSyncContext *ctx, OSyncChange *change)
    {
        osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
        plugin_environment *env = (plugin_environment *)osync_context_get_plugin_data(ctx);
        
        /*
        * Here you have to add, modify or delete a object
        * 
        */
        switch (osync_change_get_changetype(change)) {
            case CHANGE_DELETED:
                //Delete the change
                //Dont forget to answer the call on error
                if ( !env->sync->deleteItem( change ) ) {
                    osync_context_report_error(ctx, OSYNC_ERROR_IO_ERROR, "Couldn't delete from akonadi");
                    return FALSE;
                }
                break;
            case CHANGE_ADDED:
                //Add the change
                //Dont forget to answer the call on error
                if ( !env->sync->addItem( change ) ) {
                    osync_context_report_error(ctx, OSYNC_ERROR_IO_ERROR, "Couldn't add to akonadi");
                    return FALSE;
                }
                //If you are using hashtables you have to calculate the hash here:
                env->sync->updateItem( change, FALSE );
                break;
            case CHANGE_MODIFIED:
                //Modify the change
                //Dont forget to answer the call on error
                if ( !env->sync->modifyItem( change ) ) {
                    osync_context_report_error(ctx, OSYNC_ERROR_IO_ERROR, "Couldn't modify in akonadi");
                    return FALSE;
                }
                //If you are using hashtables you have to calculate the new hash here:
                env->sync->updateItem( change, FALSE );
                break;
            default:
                osync_debug("AKONADI-SYNC", 0, "Unknown change type");
        }
        //if you use hashtable, update the hash now.
        osync_hashtable_update_hash(env->hashtable, change);
        
        //Answer the call
        osync_context_report_success(ctx);
        
        osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
        return TRUE;
    }

    static void sync_done(OSyncContext *ctx)
    {
        osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
        plugin_environment *env = (plugin_environment *)osync_context_get_plugin_data(ctx);
        
        /*
        * This function will only be called if the sync was successfull
        */
        
        //If we have a hashtable we can now forget the already reported changes
        osync_hashtable_forget(env->hashtable);
        
        //If we use anchors we have to update it now.
        const char *lanchor = "akonadi-sync";
        //Now you get/calculate the current anchor of the device
        osync_anchor_update(env->member, "lanchor", lanchor);
        
        //Answer the call
        osync_context_report_success(ctx);
        osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
        return;
    }

    static void disconnect(OSyncContext *ctx)
    {
        osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
        plugin_environment *env = (plugin_environment *)osync_context_get_plugin_data(ctx);
        
        //Close all stuff you need to close
        
        //Close the hashtable
        osync_hashtable_close(env->hashtable);
        
        //Answer the call
        osync_context_report_success(ctx);
        osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
        return;
    }

    static void finalize(void *data)
    {
        osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
        
        plugin_environment *env = (plugin_environment *)data;
        
        //Free all stuff that you have allocated here.
        osync_hashtable_free(env->hashtable);
        
        if (env->sync)
            delete env->sync;
        
        osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
    }

    KDE_EXPORT void get_info(OSyncEnv *env)
    {
        osync_trace(TRACE_ENTRY, "%s", __PRETTY_FUNCTION__);
        //Now you can create a new plugin information and fill in the details
        //Note that you can create several plugins here
        OSyncPluginInfo *info = osync_plugin_new_info(env);
        
        //Tell opensync something about your plugin
        info->name = "akonadi-sync";
        info->longname = "Plugin for KDE4 PIM";
        info->description = "Sync with the Akonadi PIM framework used in KDE4.5+";
        //the version of the api we are using, (1 at the moment)
        info->version = 1;
        //If you plugin is threadsafe.
        info->is_threadsafe = TRUE;
        
        //Now set the function we made earlier
        info->functions.initialize = initialize;
        info->functions.connect = connect;
        info->functions.sync_done = sync_done;
        info->functions.disconnect = disconnect;
        info->functions.finalize = finalize;
        info->functions.get_changeinfo = get_changeinfo;
        info->config_type = OPTIONAL_CONFIGURATION;
        
        //If you like, you can overwrite the default timeouts of your plugin
        //The default is set to 60 sec. Note that this MUST NOT be used to
        //wait for expected timeouts (Lets say while waiting for a webserver).
        //you should wait for the normal timeout and return a error.
        //info->timeouts.connect_timeout = 5;
        //There are more timeouts for the other functions
        
        //Now you have to tell opensync all the object types that your are accepting.
        //There can be more than one type you accept (for example: contact, event, note etc)
        osync_plugin_accept_objtype(info, "contact");
        osync_plugin_accept_objtype(info, "event");
        osync_plugin_accept_objtype(info, "todo");
        osync_plugin_accept_objtype(info, "note");
        
        //which format do you accept for these objtypes
        osync_plugin_accept_objformat(info, "contact", "vcard30", "kde");
        osync_plugin_accept_objformat(info, "event", "vevent20", "kde");
        osync_plugin_accept_objformat(info, "todo", "vtodo20", "kde");
        osync_plugin_accept_objformat(info, "note", "vnote11", NULL);
        
        //set the commit function for these format. this function will be called for
        //each object to write once
        osync_plugin_set_commit_objformat(info, "contact", "vcard30", commit_change);
        osync_plugin_set_commit_objformat(info, "event", "vevent20", commit_change);
        osync_plugin_set_commit_objformat(info, "todo", "vtodo20", commit_change);
        osync_plugin_set_commit_objformat(info, "note", "vnote11", commit_change);
        
        //the other possibility is to do a batch commit by setting
        //osync_plugin_set_batch_commit_objformat(info, "<object type name>", "<format name>", batch_commit);
        //this function will be called exactly once with all objects to write gathered in an array
        osync_trace(TRACE_EXIT, "%s", __PRETTY_FUNCTION__);
    }
}